/*
 * These defines are byte order independent. There is no alignment of fields
 * permitted in the structures. Therefore they are declared as characters
 * and the values loaded from the character positions. It also makes it
 * nice to have it "endian" independent.
 */
 
/* Load a short int from the following tables with little-endian formats */
#define COFF_SHORT_L(ps) ((short)(((unsigned short)((unsigned char)ps[1])<<8)|\
				  ((unsigned short)((unsigned char)ps[0]))))

/* Load a long int from the following tables with little-endian formats */
#define COFF_LONG_L(ps) (((long)(((unsigned long)((unsigned char)ps[3])<<24) |\
				 ((unsigned long)((unsigned char)ps[2])<<16) |\
				 ((unsigned long)((unsigned char)ps[1])<<8)  |\
				 ((unsigned long)((unsigned char)ps[0])))))
 
/* Load a short int from the following tables with big-endian formats */
#define COFF_SHORT_H(ps) ((short)(((unsigned short)((unsigned char)ps[0])<<8)|\
				  ((unsigned short)((unsigned char)ps[1]))))

/* Load a long int from the following tables with big-endian formats */
#define COFF_LONG_H(ps) (((long)(((unsigned long)((unsigned char)ps[0])<<24) |\
				 ((unsigned long)((unsigned char)ps[1])<<16) |\
				 ((unsigned long)((unsigned char)ps[2])<<8)  |\
				 ((unsigned long)((unsigned char)ps[3])))))

/* These may be overridden later by brain dead implementations which generate
   a big-endian header with little-endian data. In that case, generate a
   replacement macro which tests a flag and uses either of the two above
   as appropriate. */

#define COFF_LONG(v)   COFF_LONG_L(v)
#define COFF_SHORT(v)  COFF_SHORT_L(v)
struct coff_filehdr {
		unsigned short f_magic;  //  
        unsigned short f_nscns;  //  
        long           f_timdat; //    
        long           f_symptr; //      
        long           f_nsyms;  //     
        unsigned short f_opthdr; //   
        unsigned short f_flags;  // 
    };

#define COFF_FILHDR  struct coff_filehdr
#define COFF_FILHSZ  sizeof (COFF_FILHDR)
/*
 *   Bits for f_flags:
 *
 *	F_RELFLG	relocation info stripped from file
 *	F_EXEC		file is executable  (i.e. no unresolved external
 *			references)
 *	F_LNNO		line numbers stripped from file
 *	F_LSYMS		local symbols stripped from file
 *	F_MINMAL	this is a minimal object file (".m") output of fextract
 *	F_UPDATE	this is a fully bound update file, output of ogen
 *	F_AR32W		32-bit word
 */
#define  COFF_F_RELFLG		0x0001
#define  COFF_F_EXEC		0x0002
#define  COFF_F_LNNO		0x0004
#define  COFF_F_LSYMS		0x0008
#define  COFF_F_LADDRS		0x0020 //App can handle > 2gb addresses.
#define  COFF_F_16BIT		0x0040 //Use of this flag is reserved for futureuse.
#define  COFF_F_REVLO		0x0080 //Little endian: LSB precedes MSB in memory.
#define  COFF_F_AR32W		0x0100
#define  COFF_F_DEBUG		0x0200 //Debugging information removed from image file.

#define	 COFF_I386MAGIC	        0x14c   

    struct scnhdr {
        char           s_name[8]; //   
        long           s_paddr;   //   
        long           s_vaddr;   //  
        long           s_size;    //  
        long           s_scnptr;  //      
        long           s_relptr;  //     .  . 
        long           s_lnnoptr; //     .   
        unsigned short s_nreloc;  //  ,  
        unsigned short s_nlnno;   //      
        long           s_flags;   //    
    };

#define SCNHDR  struct scnhdr
#define SCNHSZ  sizeof (SCNHDR)
#define COFF_STYP_REG     0x00 // regular segment 
#define COFF_STYP_DSECT   0x01 // dummy segment   
#define COFF_STYP_NOLOAD  0x02 // no-load segment 
#define COFF_STYP_GROUP   0x04 // group segment   
#define COFF_STYP_PAD     0x08 // .pad segment    
#define COFF_STYP_COPY    0x10 // copy section    
#define COFF_STYP_TEXT    0x20 // .text segment   
#define COFF_STYP_DATA    0x40 // .data segment   
#define COFF_STYP_BSS     0x80 // .bss segment    
#define COFF_STYP_INFO   0x200 // .comment section
#define COFF_STYP_OVER   0x400 // overlay section 
#define COFF_STYP_LIB    0x800 // library section 

	//MS COFF flags set for sections
#define SCN_TYPE_REG               0x00000000 //Reserved for future use.
#define SCN_TYPE_DSECT             0x00000001 //Reserved for future use.
#define SCN_TYPE_NOLOAD            0x00000002 //Reserved for future use.
#define SCN_TYPE_GROUP             0x00000004 //Reserved for future use.
#define SCN_TYPE_NO_PAD            0x00000008 //Section should not be padded to next boundary. This is obsolete and replaced by IMAGE_SCN_ALIGN_1BYTES. This is valid for object files only.
#define SCN_TYPE_COPY              0x00000010 //Reserved for future use.
#define SCN_CNT_CODE               0x00000020 //Section contains executable code.
#define SCN_CNT_INITIALIZED_DATA   0x00000040 //Section contains initialized data.
#define SCN_CNT_UNINITIALIZED_DATA 0x00000080 //Section contains uninitialized data.
#define SCN_LNK_OTHER              0x00000100 //Reserved for future use.
#define SCN_LNK_INFO               0x00000200 //Section contains comments or other information. The .drectve section has this type. This is valid for object files only.
#define SCN_TYPE_OVER              0x00000400 //Reserved for future use.
#define SCN_LNK_REMOVE             0x00000800 //Section will not become part of the image. This is valid for object files only.
#define SCN_LNK_COMDAT             0x00001000 //Section contains COMDAT data. See Section 5.5.6, COMDAT Sections, for more information. This is valid for object files only.
#define SCN_MEM_FARDATA            0x00008000 //Reserved for future use.
#define SCN_MEM_PURGEABLE          0x00020000 //Reserved for future use.
#define SCN_MEM_16BIT              0x00020000 //Reserved for future use.
#define SCN_MEM_LOCKED             0x00040000 //Reserved for future use.
#define SCN_MEM_PRELOAD            0x00080000 //Reserved for future use.
#define SCN_ALIGN_1BYTES           0x00100000 //Align data on a 1-byte boundary. This is valid for object files only.
#define SCN_ALIGN_2BYTES           0x00200000 //Align data on a 2-byte boundary. This is valid for object files only.
#define SCN_ALIGN_4BYTES           0x00300000 //Align data on a 4-byte boundary. This is valid for object files only.
#define SCN_ALIGN_8BYTES           0x00400000 //Align data on a 8-byte boundary. This is valid for object files only.
#define SCN_ALIGN_16BYTES          0x00500000 //Align data on a 16-byte boundary. This is valid for object files only.
#define SCN_ALIGN_32BYTES          0x00600000 //Align data on a 32-byte boundary. This is valid for object files only.
#define SCN_ALIGN_64BYTES          0x00700000 //Align data on a 64-byte boundary. This is valid for object files only.
#define SCN_ALIGN_128BYTES         0x00800000 //Align data on a 128-byte boundary. This is valid for object files only.
#define SCN_ALIGN_256BYTES         0x00900000 //Align data on a 256-byte boundary. This is valid for object files only.
#define SCN_ALIGN_512BYTES         0x00A00000 //Align data on a 512-byte boundary. This is valid for object files only.
#define SCN_ALIGN_1024BYTES        0x00B00000 //Align data on a 1024-byte boundary. This is valid for object files only.
#define SCN_ALIGN_2048BYTES        0x00C00000 //Align data on a 2048-byte boundary. This is valid for object files only.
#define SCN_ALIGN_4096BYTES        0x00D00000 //Align data on a 4096-byte boundary. This is valid for object files only.
#define SCN_ALIGN_8192BYTES        0x00E00000 //Align data on a 8192-byte boundary. This is valid for object files only.
#define SCN_LNK_NRELOC_OVFL        0x01000000 //Section contains extended relocations.
#define SCN_MEM_DISCARDABLE        0x02000000 //Section can be discarded as needed.
#define SCN_MEM_NOT_CACHED         0x04000000 //Section cannot be cached.
#define SCN_MEM_NOT_PAGED          0x08000000 //Section is not pageable.
#define SCN_MEM_SHARED             0x10000000 //Section can be shared in memory.
#define SCN_MEM_EXECUTE            0x20000000 //Section can be executed as code.
#define SCN_MEM_READ               0x40000000 //Section can be read.
#define SCN_MEM_WRITE              0x80000000 //Section can be written to.

struct reloc{
    long           r_vaddr;  //    
    long           r_symndx; //    
    unsigned short r_type;   //  
};
#define RELOC  reloc
#define RELSZ  sizeof(RELOC)

#define REL_ABS 0        //This relocation is ignored.
#define REL_DIR32   0x06 //The targets 32-bit virtual address.
#define REL_DIR32NB 0x07 //The targets 32-bit relative virtual address.
#define REL_SECTION 0x0A /*The 16-bit-section index of the section containing the
target. This is used to support debugging information.*/

#define REL_ABSOLUTE 0x0000 //This relocation is ignored.
#define REL_DIR16    0x0001 //Not supported.
#define REL_REL16    0x0002 //Not supported.
#define REL_DIR32    0x0006 //The targets 32-bit virtual address.
#define REL_DIR32NB  0x0007 //The targets 32-bit relative virtual address.
#define REL_SEG12    0x0009 //Not supported.
/*The 16-bit-section index of the section containing the
target. This is used to support debugging information.*/
#define REL_SECTION  0x000A
/*The 32-bit offset of the target from the beginning of its
section. This is used to support debugging information
as well as static thread local storage.*/
#define REL_SECREL   0x000B 
/*The 32-bit relative displacement to the target. This
supports the x86 relative branch and call instructions.*/
#define REL_REL32    0x0014
#define SYMNMLEN  8  //       

 struct syment {
   union {        //      
     char _n_name [SYMNMLEN]; //  
     struct {
       long _n_zeroes; //  == 0,     
       long _n_offset; //   . 
     }    _n;
 //  char *_n_nptr [2];
   }              _n;
   long           n_value;   //   
   short          n_scnum;   //  
   unsigned short n_type;    //     
   char           n_sclass;  //   
   char           n_numaux;  //    
};

#define n_name    _n._n_name
#define n_zeroes  _n._n._n_zeroes
#define n_offset  _n._n._n_offset
#define n_nptr    _n._n_nptr [1]

#define SYMENT  struct syment
#define SYMESZ  sizeof(SYMENT)

#define C_EXT 2 // 
#define C_STAT 3 //  (.text .data ...)
#define N_ABS -1 //  
#define N_UNDEF 0 //   
#define N_SCNUM 0 // >0 :  ,    